import os
import sys
import shutil
import multiprocessing
from .alignment import Parse
from .annotate import Analysis, Annotate, Map
from .translate import peptideBreak2
from os.path import basename

def main(inputfile, outputfolder, binariesfolder, databasefolder, diamondfolder):
    """Runs mifaser pipeline; is called by mifaser.py"""

    log_msgs = []

    cpu_count = multiprocessing.cpu_count()

    print("     Detected CPUs #%i" % cpu_count)

    read_file = os.path.splitext(os.path.basename(inputfile))[0]

    ## paths
    path_to_peptide_file = os.path.join(outputfolder, "peptide/")
    path_to_DIAMOND_result = os.path.join(outputfolder, "diamond/")
    path_to_parse_result = os.path.join(outputfolder, "parsed/")
    path_to_analyze_result = os.path.join(outputfolder, "analyzed/")
    path_to_readmap_dir = os.path.join(path_to_analyze_result, 'read_map')
    path_to_database_ec_file = os.path.join(databasefolder, "gold_standard_list")
    path_to_database_ec_annotation_file = os.path.join(databasefolder, "gold_standard_EC_annotation")
    diamond_binary = os.path.join(diamondfolder, 'diamond')
    database_file = os.path.join(databasefolder, "GS.dmnd")
    log_file = os.path.join(path_to_analyze_result, 'mifaser.log')

    print("     Creating output directories...")
    if not os.path.exists(outputfolder):
        os.mkdir(outputfolder)
    if not os.path.exists(path_to_peptide_file):
        os.mkdir(path_to_peptide_file)
    if not os.path.exists(path_to_DIAMOND_result):
        os.mkdir(path_to_DIAMOND_result)
    if not os.path.exists(path_to_parse_result):
        os.mkdir(path_to_parse_result)
    if os.path.exists(path_to_analyze_result):
        shutil.rmtree(path_to_analyze_result)
    os.mkdir(path_to_analyze_result)

    ## processing flags
    abort = False

    ## call the translate script
    print("     Translating reads...")
    do_translate = True
    # --- DEPRECATED ---
    # TODO: define switch to toggle translation
    # 
    # with open(inputfile) as fp:
    #     translated_file = "%s%s.prd" % (path_to_peptide_file, read_file)
    #     for line in fp:
    #         if line.find('NO_TRANSLATE;', 0, 15) != -1:
    #             print("     Found (NO_TRANSLATE;): skipping translation.")
    #             do_translate = False
    #             os.rename(inputfile, translated_file)
    #             break
    # -----------------
    if do_translate:
       peptideBreak2.main(read_file, inputfile, path_to_peptide_file)

    ## assert successfull translation
    prd_file = '%s%s.prd' % (path_to_peptide_file, read_file)
    if not os.path.exists(prd_file):
        abort = True
        log_msgs.append('ERROR#1 - No translated peptide file found. Aborting.')
    else:
        with open(prd_file) as peptide_file:
            first = peptide_file.read(1)
            if not first:
                abort = True
                log_msgs.append('ERROR#2 - Translated peptide file is empty. Aborting.')
    if abort:
        analysis=open(os.path.join(path_to_analyze_result, 'analysis.tsv'),"w")
        analysis.write('0\t0\n')
        analysis.close()
        ec_count=open(os.path.join(path_to_analyze_result, 'ec_count.tsv'),"w")
        ec_count.close()
        multi_ec=open(os.path.join(path_to_analyze_result, 'multi_ec.tsv'),"w")
        multi_ec.close()
        os.mkdir(path_to_readmap_dir)
        with open(log_file, "w+")as mifaser_log:
            mifaser_log.writelines("%s\n" % l for l in log_msgs)
        print(" >>> aborted: ERROR. logs: %s" % os.path.join(path_to_analyze_result, 'mifaser.log'))
        sys.exit(0)

    ##################################
    # TODO implement multi-threading #
    ##################################

    ## run DIAMOND with the tranlslated peptide file
    print("     Running DIAMOND...")
    cmd =  "%s blastp -p %i -q %s%s.prd -d %s -k 1000000 --min-score 10 -a %s%s --quiet" % (diamond_binary, cpu_count, path_to_peptide_file, read_file, database_file, path_to_DIAMOND_result, read_file)
    os.system(cmd)

    ## convert DIAMOND result from .daa to .tab
    print("     Converting DIAMOND...")
    cmd = "%s view -p %i -a %s%s.daa -o %s%s.out -f tab --quiet" % (diamond_binary, cpu_count, path_to_DIAMOND_result, read_file, path_to_DIAMOND_result, read_file)
    os.system(cmd)

    ## parse the DIAMOND result
    print("     Parsing DIAMOND...")
    Parse.main(read_file, path_to_DIAMOND_result, path_to_parse_result)

    ## Annotate parsed result with ECs.
    print("     Annotating...")
    Annotate.main(read_file, path_to_parse_result, path_to_database_ec_file, path_to_analyze_result)

    ## Analysis the data, exclude multiEC and get final EC, read_count table.
    print("     Analyzing...")
    Analysis.main(read_file, path_to_analyze_result, path_to_database_ec_annotation_file)

    ## Generate a multi-fasta file of reads that are mapped to the ECs
    print("     Create read mapping...")
    Map.main(path_to_analyze_result, inputfile)

    print("   - finished")
