import argparse
import re
from . import Gene

"""" This script will take a .rds file and translate every read. It will then
return all amino acid sequences of length >= 11. AA sequences are broken up
first by reading frame, then by stop codon.

By Pavel Vaysberg
pvaysberg@gmail.com
Last updated: 4/9/14
"""

# DNA codon table
bases = ['T', 'C', 'A', 'G']
codons = [a+b+c for a in bases for b in bases for c in bases]
amino_acids = 'FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG'
codon_table = dict(zip(codons, amino_acids))


# Function to translate a DNA sequence, looking at all 6 reading frames
# Will return a list of all AA sequences >= 11, separated by stop codons

class Peptide:
    chain = ""
    genes = []

    def __init__(self, chain):
        self.chain = str(chain)
        self.genes = []
    def addgene(self, gene):
        self.genes.append(gene)
        return
    def getgenenames(self):
        names = []
        for gene in self.genes:
            names.append(gene.name)
        return names

def replace(sequence):
    temp = ""
    for x in sequence:

        if x == 'A' :
            temp += 'T'
            continue
        if x == 'T' :
            temp += 'A'
            continue
        if x == 'C' :
            temp += 'G'
            continue
        if x == 'G' :
            temp += 'C'
            continue
    newsequence = temp
    return newsequence

#####By C.S., it trim the reads on both sides, removing the "n"s
# def trim(sequence):
#   if "N" in sequence:
#       mid = int(len(sequence)/2)
#       seq_1 = sequence[:mid]
#       seq_2 = sequence[mid:]
#       index_1 = seq_1.rfind("N")
#       seq_1_trim = seq_1[(index_1+1):]
#       index_2 = seq_2.find("N")
#       seq_2_trim = seq_2[:index_2]
#       seq_trim = seq_1_trim + seq_2_trim
#   else:
#       seq_trim = sequence
#   return seq_trim

#####By C.S.M.M., it trim the reads on both sides, removing all alternatives
def trim(sequence):
    alternatives_zero =["Z"]
    alternatives_two = ["W", "S", "M", "K", "R", "Y"]
    alternatives_three = ["B", "D", "H", "V"]
    alternatives_four = ["N"]
    alternatives_all = alternatives_zero + alternatives_two + alternatives_three + alternatives_four
    mid = int(len(sequence)/2)
    seq_1 = sequence[:mid]
    seq_2 = sequence[mid:]
    index_left = 0
    index_right = len(seq_2)
    for alterative in alternatives_all:
        if alterative in sequence:
            # mid = int(len(sequence)/2)
            # seq_1 = sequence[:mid]
            # seq_2 = sequence[mid:]
            index_1 = seq_1.rfind(alterative)
            if (index_1 + 1) > index_left:
                index_left = index_1 + 1
            #seq_1_trim = seq_1[(index_1+1):]
            index_2 = seq_2.find(alterative)
            if index_2 != -1 and index_2 < index_right:
                index_right = index_2
            #seq_2_trim = seq_2[:index_2]
            #seq_trim = seq_1_trim + seq_2_trim
            #else:
            #   seq_trim = sequence
        #print(index_left, index_right)
    seq_1_trim = seq_1[index_left:]
    seq_2_trim = seq_2[:index_right]
    seq_trim = seq_1_trim + seq_2_trim
    return seq_trim


def translate(sequence, genes):


    peptidelist = [] # return value

    for s in range(0,2): # 2 strands
        for n in range(0,3): #  3 reading frames

            #sequence1 = sequence.strip().upper()
            sequence1 = trim(sequence.strip().upper())

            if s == 1:  # opposite strand
                sequence1 = replace(sequence1[::-1])

            presequence = ""
            for c in range(0,len(sequence),3):

                codon = sequence1[c+n:(c+n)+3]
                if len(codon) < 3: #we've reached the end
                    break



                presequence = presequence + codon_table[codon] # building our peptide sequence


            fraglist = presequence.split('*')

            pointer = 1 # this pointer tracks the location in the gene that our slices of AA sequences correspond to

            for x in fraglist: # assign genes to each peptide

                if not pointer == 1: pointer += 1 # accounting for stop codon

                if len(x) < 11:
                    if x == '': pointer += 1 # this accounts for multiple stop codons
                    pointer += len(x)
                    continue
                else:
                    peptide = Peptide(x)


                for gene in genes:
                    if ((gene.strand == '+' and s == 1) or (gene.strand == '-' and s == 0)): # gene is in opposite strand, don't need it
                           continue

                    pepstart = 1 + 3*(pointer-1)
                    pepend = (pointer + len(x))*3

                    if (s == 1 and gene.strand == '-'): # looking for gene on opposite strand
                        mstrand_start = len(sequence1) - gene.locOnReadEnd
                        mstrand_end = len(sequence1) - gene.locOnReadStart

                        if ((( pepstart + n ) > mstrand_end) or ( pepend + n < mstrand_start)): # checks if the amino acid sequence is NOT in range of the gene
                            continue
                        else:
                            peptide.addgene(gene)

                    elif ((( pepstart + n ) > gene.locOnReadEnd) or ( pepend + n < gene.locOnReadStart)): # checks if the amino acid sequence is NOT in range of the gene
                        continue
                    else:
                        peptide.addgene(gene)


                pointer += len(x)
                peptidelist.append(peptide)
    return peptidelist


def main(organism, rdsfile, result_path):
    #name = rdsfile.replace('.txt','')
    #name = name[:-4]

    outfile = open(result_path + organism + '.prd', 'w')

    #outfile = open(name + '.prd', 'w')
    with open(rdsfile, 'r') as reads:
        for line in reads:
            if '>' in line:
               header = line.replace('\n', '')
               header = header.replace('>', '')
               header = header.replace(' ', '')
               readline = header.split(',')
               sequence = next(reads)
               genes = []

               for n in range(3,len(readline)-1,3):
                    locsonread = readline[n+2].split('-')
                    genes.append(Gene.Gene(readline[n], readline[n+1], int(locsonread[0]), int(locsonread[1])))

               peptides = translate(sequence, genes)
               count = 1
               for peptide in peptides:
                   outfile.write('>' + ','.join(readline) + ',' + 'P_' + str(count) + ',' + ','.join(map(str, peptide.getgenenames())) + '\n' + peptide.chain + '\n')
                   count += 1
    outfile.close()
    reads.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser("""This script will take an .rds file and translate the sequences within.It will return an output file containing all peptides of length >= 11""")

    parser.add_argument('-o', '--org', help = "The metagenome, format: metagenomeid_lane")
    parser.add_argument('-r', '--rds', help = "The .rds file you'd like to translate (include path)")
    parser.add_argument('-p', '--pth', help = "The path to store the result file (include the final '/').")

    args = parser.parse_args()
    rdsfile = args.rds
    organism = args.org
    result_path = args.pth

    if not organism:
        organism = raw_input("Enter an metagenome id , format: metagenomeid_lane: ")

    if not rdsfile:
        rdsfile = raw_input("Enter an .rds file (include path): ")

    if not result_path:
        result_path = raw_input("The path to store the result file (include the final '/').")

    main(organism, rdsfile, result_path)
