import sys
if sys.version_info[0] < 3:
    raise Exception("mi-faser requires Python 3.x - please install the appropriate version.")
import os
import subprocess
import shutil
import gzip
import argparse
import csv
import datetime
from binaries import Caller
from collections import OrderedDict


VERSION_ID = '1.11.3'
VERSION_DATE = '11/13/17'


def file_len(fname):
    i = -1
    with open(fname) as f_:
        for i, l in enumerate(f_):
            pass
    return i + 1


def fasta_header_count(fname):
    header_count = 0
    with open(fname) as f_:
        for i, l in enumerate(f_):
            if l[0] == ">":
                header_count += 1
    return header_count


def split_by_length(fastafile, outfolder, splitlength):
    default_separator = '_'
    prefix = ''
    separator = ''

    split_idx = 1
    out_files = []
    total_splitted = 0
    input_file = open(fastafile, 'r')
    split_out_base = outfolder
    if prefix:
        split_out_base = os.path.join(split_out_base, prefix)
    else:
        split_out_base = os.path.join(split_out_base, os.path.splitext(os.path.basename(fastafile))[0])
    if separator:
        split_out_base += separator
    else:
        split_out_base += default_separator

    reading = True
    while reading:
        first = True
        split_out_filename = "%s%s.fasta" % (split_out_base, split_idx)
        output_file = open(split_out_filename, 'w')
        out_files.append(split_out_filename)
        current_splitted = 0
        while current_splitted < splitlength:
            line = input_file.readline()
            while not line.strip():
                line = input_file.readline()
                if len(line) == 0:
                    break
            header = line
            current_splitted += 1
            sequence = input_file.readline().rstrip()
            if not header.rstrip() and not sequence:
                reading = False
                if first:
                    output_file.close()
                    os.remove(split_out_filename)
                    split_idx -= 1
                    out_files.pop(-1)
                break
            last_pos = input_file.tell()
            line = input_file.readline().rstrip()
            while line and line[0] != '>':
                sequence += line
                last_pos = input_file.tell()
                line = input_file.readline().rstrip()
            if line and line[0] == '>':
                input_file.seek(last_pos)
            output_file.write(header + sequence + "\n")
            total_splitted += 1
            first = False
        output_file.close()
        split_idx += 1
    return out_files, total_splitted


def find_fasta_in_dir(root_dir):
    found_files = []
    for root, directories, filenames in os.walk(root_dir):
        for filename in filenames:
            if filename[0] == '.':
                continue
            elif filename.endswith(".fasta") or filename.endswith(".fa") or filename.endswith(".fna") \
                    or filename.endswith(".fastq") or filename.endswith(".fq"):
                found_files.append(os.path.join(root, filename))
    return found_files


def preprocess_input(file, split, outdir, splits):
    """Preprocessing: split input files"""

    input_file_name = os.path.basename(file)
    input_file_base = os.path.splitext(input_file_name)[0]

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    # check if file is compressed: zip, gz, tar, tgz
    is_compressed = True
    input_file_tmp_extract_dir = os.path.join(outdir, input_file_name)
    if input_file_name.endswith('.zip'):
        print("     Extracting zip file %s... " % input_file_name)
        os.mkdir(input_file_tmp_extract_dir)
        shutil.unpack_archive(file, input_file_tmp_extract_dir, 'zip')
    elif input_file_name.endswith('.bzip2'):
        print("     Extracting bzip2 file %s... " % input_file_name)
        os.mkdir(input_file_tmp_extract_dir)
        shutil.unpack_archive(file, input_file_tmp_extract_dir, 'bztar')
    elif input_file_name.endswith('.xz'):
        print("     Extracting xz file %s... " % input_file_name)
        os.mkdir(input_file_tmp_extract_dir)
        shutil.unpack_archive(file, input_file_tmp_extract_dir, 'xztar')
    elif input_file_name.endswith('.tar'):
        print("     Extracting tar file %s... " % input_file_name)
        os.mkdir(input_file_tmp_extract_dir)
        shutil.unpack_archive(file, input_file_tmp_extract_dir, 'tar')
    elif input_file_name.endswith('.tar.gz'):
        print("     Extracting targz file %s... " % input_file_name)
        os.mkdir(input_file_tmp_extract_dir)
        shutil.unpack_archive(file, input_file_tmp_extract_dir, 'gztar')
    elif input_file_name.endswith('.gz') or input_file_name.endswith('.gzip'):
        print("     Extracting gz/gzip file %s... " % input_file_name)
        os.mkdir(input_file_tmp_extract_dir)
        input_file_name = input_file_name.rsplit('.', 1)[0]
        with gzip.open(file, 'rb') as f_in,\
                open(os.path.join(input_file_tmp_extract_dir, input_file_name), 'wb') as f_out:
            f_out.writelines(f_in)
    elif input_file_name.endswith('.bz2') or input_file_name.endswith('.bzip2'):
        print("     Extracting bz2/bzip2 file %s... " % input_file_name)
        try:
            import bz2
        except ImportError:
            print("       ERROR: bz2 format not supported: could not import bz2 module")
            raise
        os.mkdir(input_file_tmp_extract_dir)
        input_file_name = input_file_name.rsplit('.', 1)[0]
        with bz2.BZ2File(file, 'rb') as f_in, \
                open(os.path.join(input_file_tmp_extract_dir, input_file_name), 'wb') as f_out:
            f_out.writelines(f_in)
    else:
        is_compressed = False
        print(" *** Input: Uncompressed file %s" % input_file_name)
    if is_compressed:
        extracted_fasta_files = find_fasta_in_dir(input_file_tmp_extract_dir)
        print(" *** Extracted %i files, calling mifaser for each of them." % len(extracted_fasta_files))
        for extracted_file in extracted_fasta_files:
            preprocess_input(extracted_file, split, outdir, splits)
        return

    input_format = 'fasta'
    if input_file_name.endswith('.fq') or input_file_name.endswith('.fastq'):
        input_format = 'fastq'

    if input_format == 'fastq':
        try:
            from Bio import SeqIO
            print("     Converting input file from fastq to fasta format: %s... " % input_file_name)
            input_file_converted = os.path.join(outdir, '%s.fasta' % input_file_base)
            SeqIO.convert(file, "fastq", input_file_converted, "fasta")
            file = input_file_converted
        except ImportError:
            raise Exception("The SeqIO module from the biopython package is required to handle the fastq format.")

    print("     Getting file statistics... ")
    file_size_mb = os.path.getsize(file) / 1000000
    print("     Input file size: %iMB" % file_size_mb)

    if split > 0:
        print("     Splitting input file... ")
        infile_list, read_count = split_by_length(fastafile=file, outfolder=outdir, splitlength=split)
        print(" *** Running mifaser on %i splits" % len(infile_list))
    else:
        infile_list = [file]
        # print("Getting read count... ")
        # read_count = fasta_header_count(file)
        print("     Running mifaser on single input")

    splits.extend(infile_list)


def postprocess_results(splits, splits_out, outdir):
    """Merge results into one single file"""

    total_jobcount = len(splits)

    resultfolders_list = []
    for file in splits:
        file_name = os.path.splitext(os.path.basename(file))[0]
        job_folder_path = os.path.join(os.path.abspath(splits_out), file_name)
        if os.path.isdir(job_folder_path):
            resultfolders_list.append(job_folder_path)
        else:
            print(" WARNING: result folder %s not found" % (job_folder_path))
    print(" *** Merging %i splits..." % len(resultfolders_list))

    if total_jobcount != len(resultfolders_list):
        print(" WARNING: found %i result folders for %i jobs" % (len(resultfolders_list), total_jobcount))

    merged_result = {}
    merged_ec_count = []
    merged_multi_ec = {}
    merged_stats = {'ec_annotated_total': 0, 'ec_annotated_multiple': 0}

    missing_analysis_list = []
    missing_joblist = []
    missing_eccount_list = []
    missing_multiec_list = []

    for output_folder in resultfolders_list:
        analysis_file = os.path.join(output_folder, 'analyzed/analysis.tsv')
        eccount_file = os.path.join(output_folder, 'analyzed/ec_count.tsv')
        multiec_file = os.path.join(output_folder, 'analyzed/multi_ec.tsv')

        # analysis merge
        if not os.path.exists(analysis_file):
            missing_analysis_list.append(output_folder)
            missing_job_full_id = output_folder.split("/")[-1]
            missing_job_jobid = int(missing_job_full_id.split('.')[0])
            missing_job_arrayidx = int(missing_job_full_id.split('.')[1])
            missing_joblist.append((missing_job_jobid, missing_job_arrayidx))
            print(" WARNING: Missing analysis file: %s" % analysis_file)
        else:
            with open(analysis_file, 'r') as tsv_in:
                tsv_in = csv.reader(tsv_in, delimiter='\t')

                # # header statistics: ec_annotated_total  ec_annotated_multiple
                # old_header_stats = next(tsv_in)
                # header_stats = [old_header_stats[0].split()[5], old_header_stats[0].split()[10]]
                header_stats = next(tsv_in)
                merged_stats['ec_annotated_total'] = merged_stats['ec_annotated_total'] + int(header_stats[0])
                merged_stats['ec_annotated_multiple'] = merged_stats['ec_annotated_multiple'] + int(header_stats[1])

                for row in tsv_in:
                    _ec = row[0]
                    _readcount = int(row[1])
                    # description = row[2]

                    if _ec not in merged_result.keys():
                        merged_result[_ec] = _readcount
                    else:
                        merged_result[_ec] = merged_result[_ec] + _readcount

        # ec_count merge
        if not os.path.exists(eccount_file):
            print(" WARNING: Missing eccount file: %s" % eccount_file)
            missing_eccount_list.append(output_folder)
        else:
            with open(eccount_file, 'r') as tsv_in:
                tsv_in = csv.reader(tsv_in, delimiter='\t')
                for row in tsv_in:
                    merged_ec_count.append(row)

        # multiec merge
        if not os.path.exists(multiec_file):
            print(" WARNING: Missing multiec file: %s" % multiec_file)
            missing_multiec_list.append(output_folder)
        else:
            with open(multiec_file, 'r') as tsv_in:
                tsv_in = csv.reader(tsv_in, delimiter='\t')
                for row in tsv_in:
                    _ecs = row[0]
                    _readcounts = int(row[1])
                    if _ecs not in merged_multi_ec.keys():
                        merged_multi_ec[_ecs] = _readcounts
                    else:
                        merged_multi_ec[_ecs] = merged_multi_ec[_ecs] + _readcounts

    merged_result = OrderedDict(sorted(merged_result.items()))

    # ec read mapping merge
    merged_ec_mapping_dir = os.path.join(outdir, 'read_map')
    if not os.path.exists(merged_ec_mapping_dir):
        os.mkdir(merged_ec_mapping_dir)
    for ec in merged_result.keys():
        merged_ec_mapping_file = os.path.join(merged_ec_mapping_dir, "%s.fasta" % ec)
        with open(merged_ec_mapping_file, 'w') as outfile:
            for output_folder in resultfolders_list:
                ec_mapping_dir = os.path.join(output_folder, 'analyzed/read_map')
                ec_mapping_file = os.path.join(ec_mapping_dir, "%s.fasta" % ec)
                if os.path.exists(ec_mapping_file):
                    with open(ec_mapping_file, 'r') as infile:
                        outfile.write(infile.read())

    result_file = os.path.join(outdir, 'analysis.tsv')
    with open(result_file, 'w') as tsv_out:
        tsv_writer = csv.writer(tsv_out, delimiter='\t')
        tsv_writer.writerow([merged_stats['ec_annotated_total'], merged_stats['ec_annotated_multiple']])
        for ec, readcount in merged_result.items():
            tsv_writer.writerow([ec, readcount])

    result_eccount = os.path.join(outdir, 'ec_count.tsv')
    with open(result_eccount, 'w') as tsv_out:
        tsv_writer = csv.writer(tsv_out, delimiter='\t')
        for row in merged_ec_count:
            tsv_writer.writerow(row)

    result_multiec = os.path.join(outdir, 'multi_ec.tsv')
    with open(result_multiec, 'w') as tsv_out:
        tsv_writer = csv.writer(tsv_out, delimiter='\t')
        for ecs, count in merged_multi_ec.items():
            tsv_writer.writerow([ecs, count])

    error_file = os.path.join(outdir, 'errors.txt')
    if len(missing_analysis_list) or len(missing_eccount_list) or len(missing_multiec_list) \
            or (total_jobcount != len(resultfolders_list)):
        msg = "Missing result files -> analysis: %i, eccount: %i, multiec: %i, resultfolders/jobs: %i/%i" \
              % (len(missing_analysis_list),
                 len(missing_eccount_list),
                 len(missing_multiec_list),
                 len(resultfolders_list),
                 total_jobcount)
        print(" WARNING: " + msg)
        with open(error_file, 'w') as tsv_out:
            tsv_writer = csv.writer(tsv_out, delimiter='\t')
            tsv_writer.writerow(['# ' + msg])
            # tsv_writer.writerow(['# all retrieved'])
            # tsv_writer.writerow(resultfolders_list)
            tsv_writer.writerow(['# missing_analysis_list'])
            tsv_writer.writerow(missing_analysis_list)
            tsv_writer.writerow(['# missing_eccount_list'])
            tsv_writer.writerow(missing_eccount_list)
            tsv_writer.writerow(['# missing_multiec_list'])
            tsv_writer.writerow(missing_multiec_list)


def main(inputfile, outputfolder, binariesfolder, databasefolder, diamondfolder, splitsize, platform):
    start = datetime.datetime.now()
    print("\n --- mi-faser %s (%s) ---" % (VERSION_ID, VERSION_DATE))

    file_name = os.path.splitext(os.path.basename(inputfile))[0]

    if not outputfolder:
        outputfolder_user = '%s_out' % file_name
    else:
        outputfolder_user = outputfolder

    splits_folder = os.path.join(outputfolder_user, 'splits')

    diamond_version = subprocess.check_output(
        ['%s/diamond/%s/diamond' % (os.path.dirname(os.path.realpath(__file__)), platform), '--version'],
        shell=False
    ).decode("utf-8").replace('diamond version', '').strip()

    print("   - [ DIAMOND v%s ] -\n" % diamond_version)

    splits = []
    preprocess_input(inputfile, splitsize, splits_folder, splits)

    counter = 1
    all_splits = len(splits)
    for file in splits:
        print("   * %i/%i | %s" % (counter, all_splits, os.path.basename(file)))
        file_name = os.path.splitext(os.path.basename(file))[0]
        Caller.main(os.path.abspath(file),
                    os.path.join(os.path.abspath(splits_folder), file_name),
                    os.path.abspath(binariesfolder),
                    os.path.abspath(databasefolder),
                    os.path.abspath(diamondfolder))
        counter += 1

    postprocess_results(splits, splits_folder, outputfolder_user)

    print(" *** Results saved to: %s" % os.path.abspath(outputfolder_user))

    end = datetime.datetime.now()
    delta = end-start
    print("\n --- finished in %s --- \n" % str(delta))


if __name__ == "__main__":
    cwd = os.path.abspath(os.path.dirname(sys.argv[0]))
    platform_ = 'na'
    if sys.platform == "linux" or sys.platform == "linux2":
        platform_ = 'linux'
    elif sys.platform == "darwin":
        platform_ = 'osx'
    elif sys.platform == "win32" or sys.platform == "win64":
        platform_ = 'win'

    parser = argparse.ArgumentParser("""mi-faseri, microbiome - functional annotation of sequencing reads""")

    parser.add_argument('-f', '--inputfile',
                        help="the input DNA reads file (include path)")
    parser.add_argument('-o', '--outputfolder',
                        help="the path of the base output folder")
    parser.add_argument('-b', '--binariesfolder', default=os.path.join(cwd, 'binaries'),
                        help="the path to the folder containing all binaries and scripts")
    parser.add_argument('-d', '--databasefolder', default=os.path.join(cwd, 'database'),
                        help="the path to the folder containing all database files")
    parser.add_argument('-i', '--diamondfolder', default=os.path.join(cwd, os.path.join('diamond', platform_)),
                        help="the path to the folder containing the diamond binary")
    parser.add_argument('-s', '--split', default=100000, type=int,
                        help="splitsize to partition input")

    args = parser.parse_args()
    inputfile_ = args.inputfile
    outputfolder_ = args.outputfolder
    binariesfolder_ = args.binariesfolder
    databasefolder_ = args.databasefolder
    diamondfolder_ = args.diamondfolder
    splitsize_ = args.split

    if not inputfile_:
        inputfile_ = input("Enter the input DNA reads file: ")

    inputfile_ = os.path.abspath(inputfile_)
    if not os.path.exists(inputfile_):
        print("ERROR: Could not read input file: %s" % inputfile_)
        exit(1)

    main(inputfile_, outputfolder_, binariesfolder_, databasefolder_, diamondfolder_, splitsize_, platform_)
